import search

# Assignment problem (as a cost minimization problem)

class AssignmentProblem(search.Problem):
    
    def __init__(self, costs: list[list[float]]):
        self.costs = costs

    def __str__(self):
        return 'Cost matrix '+str(self.costs)

    def size(self) -> int:
        return len(self.costs) # == len(self.weights)



class AssignmentSolution(search.Solution):
    def __init__(self, problem: AssignmentProblem, assignment: list[int] = []):
        self.problem = problem
        self.assignment = assignment

    def __str__(self):
        return 'assignment '+str(self.assignment) + ' cost='+str(self.eval())+' bounds=['+str(self.lower_bound())+','+str(self.upper_bound())+']'

    def size(self):
        return len(self.assignment)

    def eval(self) -> float:
        cost = 0
        for i in range(self.size()):
            cost += self.problem.costs[self.assignment[i]][i]
        return cost

    def solution(self) -> bool:
        return self.size() == self.problem.size()

    def children(self) -> list[search.Solution]:
        nodes = []

        if not self.solution():
           for i in range(self.problem.size()):
               if not i in self.assignment:
                   node = AssignmentSolution(self.problem, self.assignment.copy()+[i])
                   nodes.append(node)

        return nodes


    def lower_bound(self) -> float:
        bound = self.eval() # Decisions already made
        for i in range(self.size(), self.problem.size()):
            min_alternative = self.problem.costs[0][i]
            for j in range(1,self.problem.size()):
                if self.problem.costs[j][i]<min_alternative:
                   min_alternative = self.problem.costs[j][i]
            bound += min_alternative
        return bound

    def estimation(self) -> float:
        return self.lower_bound()


    def greedy(self) -> search.Solution:
        assignment = self.assignment.copy()
        for i in range(self.size(), self.problem.size()):
            best = float("inf")
            best_assignment = None
            for j in range(self.problem.size()):
                if not j in assignment:
                   cost = self.problem.costs[j][i]
                   if cost<best:
                       best = cost
                       best_assignment = j
            assignment += [best_assignment]
        return AssignmentSolution(self.problem, assignment)

    def upper_bound(self) -> float:
        greedy_solution = self.greedy()
        return greedy_solution.eval()

    def upper_bound_solution(self) -> (float, search.Solution):
        greedy_solution = self.greedy()
        return (greedy_solution.eval(), greedy_solution)


if __name__=='__main__':

    print("ASSIGNMENT PROBLEM")
    problem = AssignmentProblem([[94,1,54,68],[74,10,88,82],[62,88,8,76],[11,74,81,21]])
    print(problem)

    root = AssignmentSolution(problem)

    print("GREEDY")
    print(root.greedy())

    print("BRANCH & BOUND")
    (c,s) = search.branch_and_bound_min(root)
    print(f"Best cost = {c}")
    print(f"Best solution = {s}")

    print("BRANCH & BOUND (no greedy)")
    (c,s) = search.branch_and_bound_min_no_greedy(root)
    print(f"Best cost = {c}")
    print(f"Best solution = {s}")


    print("ASSIGNMENT PROBLEM MAX (COSTS = -BENEFITS)")
    problem = AssignmentProblem([[-5,-6,-4],[-3,-8,-2],[-6,-5,-1]])
    print(problem)

    root = AssignmentSolution(problem)

    print("GREEDY")
    print(root.greedy())

    print("BRANCH & BOUND")
    (c,s) = search.branch_and_bound_min(root)
    print(f"Best cost = {c}")
    print(f"Best solution = {s}")

    print("BRANCH & BOUND (no greedy)")
    (c,s) = search.branch_and_bound_min_no_greedy(root)
    print(f"Best cost = {c}")
    print(f"Best solution = {s}")


