import search

# Knapsack problem

class KnapsackProblem(search.Problem):
    
    def __init__(self, capacity: float, weights: list[float], benefits: list[float]):
        self.capacity = capacity
        self.weights = weights
        self.benefits = benefits

    def __str__(self):
        return f'Size {self.capacity}\nWeights {self.weights}\nBenefits {self.benefits}'

    def size(self) -> int:
        return len(self.weights) # == len(self.benefits)


class KnapsackSolution(search.Solution):
    def __init__(self, problem: KnapsackProblem, assignment: list[int] = []):
        self.problem = problem
        self.assignment = assignment

    def __str__(self):
        return f'assignment {self.assignment} weight={self.weight()} benefit={self.eval()} est={self.estimation():.2f} bounds=[{self.lower_bound():.2f},{self.upper_bound():.2f}] '

    def __lt__(self, other): 
        '''
        Node comparison for the priority queue (maximization problem)
        '''
        return self.estimation()>other.estimation()

    def size(self):
        return len(self.assignment)

    def eval(self) -> float:
        benefit = 0
        for i in range(self.size()):
            benefit += self.assignment[i]*self.problem.benefits[i]
        return benefit

    def weight(self) -> float:
        total = 0
        for i in range(self.size()):
            total += self.assignment[i]*self.problem.weights[i]
        return total

    def solution(self) -> bool:
        return self.size() == self.problem.size()

    def children(self) -> list[search.Solution]:
        nodes = []

        if not self.solution():
            if self.weight()<=self.problem.capacity:
                node = KnapsackSolution(self.problem, self.assignment.copy()+[0])
                nodes.append(node)
            if self.weight()+self.problem.weights[self.size()]<=self.problem.capacity:
                node = KnapsackSolution(self.problem, self.assignment.copy()+[1])
                nodes.append(node)

        return nodes


    def greedy(self) -> (search.Solution, float):
        # Problema de la mochila fraccional
        assignment = self.assignment.copy()

        remaining = list(range(self.size(), self.problem.size()))

        for object in remaining:
            assignment += [0]

        remaining.sort(key=lambda x: (self.problem.benefits[x]/self.problem.weights[x]), reverse=True)    
        space = self.problem.capacity - self.weight()
        benefit = self.eval()

        for object in remaining:

            if self.problem.weights[object]<=space:
                space -= self.problem.weights[object]
                benefit += self.problem.benefits[object]
                assignment[object] = 1
            else: # Fracational value
                benefit += self.problem.benefits[object] * space / self.problem.weights[object]
                space = 0

        return (KnapsackSolution(self.problem, assignment), benefit)

 
    def lower_bound(self) -> float:
        return self.eval()

    def estimation(self) -> float:
        greedy_solution, greedy_value = self.greedy()
        return greedy_solution.eval()

    def upper_bound(self) -> float:
        greedy_solution, greedy_value = self.greedy()
        return greedy_value

    def lower_bound_solution(self) -> (float, search.Solution):
        greedy_solution, greedy_value = self.greedy()
        return (greedy_solution.eval(), greedy_solution)


if __name__=='__main__':

    print("KNAPSACK PROBLEM M=61")
    problem = KnapsackProblem(61, [1,11,21,23,33], [11,21,31,33,43])
    print(problem)

    root = KnapsackSolution(problem)

    print("Greedy solution")
    greedy_solution, greedy_value = root.greedy()
    print(greedy_solution)

    print("Branch & bound solution")
    (c,s) = search.branch_and_bound_max(root)
    print(f"Best cost = {c}")
    print(f"Best solution = {s}")

    print("Branch & bound solution (no greedy)")
    (c,s) = search.branch_and_bound_max_no_greeedy(root)
    print(f"Best cost = {c}")
    print(f"Best solution = {s}")


    print("KNAPSACK PROBLEM M=7")
    problem = KnapsackProblem(7, [1,2,3,4], [2,3,4,5])
    print(problem)

    root = KnapsackSolution(problem)

    print("Greedy solution")
    greedy_solution, greedy_value = root.greedy()
    print(greedy_solution)

    print("Branch & bound solution")
    (c,s) = search.branch_and_bound_max(root)
    print(f"Best cost = {c}")
    print(f"Best solution = {s}")

    print("Branch & bound solution")
    (c,s) = search.branch_and_bound_max_no_greeedy(root)
    print(f"Best cost = {c}")
    print(f"Best solution = {s}")
