import search

# Ski problem

class SkiProblem(search.Problem):
    
    def __init__(self, heights: list[float], lengths: list[float]):
        self.heights = heights
        self.lengths = lengths

    def __str__(self):
        return 'Heights '+str(self.heights)+'\nLengths '+str(self.lengths)

    def size(self) -> int:
        return len(self.heights) # == len(self.weights)


# Ski solution

class SkiSolution(search.Solution):

    def __init__(self, problem: SkiProblem, assignment: list[int] = []):
        self.problem = problem
        self.assignment = assignment

    def __str__(self):
        return 'assignment '+str(self.assignment) + ' cost='+str(self.eval())+' bounds=['+str(self.lower_bound())+','+str(self.upper_bound())+']'

    def size(self):
        return len(self.assignment)

    def eval(self) -> float:
        cost = 0
        for i in range(self.size()):
            cost += abs( self.problem.lengths[self.assignment[i]] - self.problem.heights[i] )
        return cost

    def solution(self) -> bool:
        return self.size() == self.problem.size()

    def children(self) -> list[search.Solution]:
        nodes = []

        if not self.solution():
           for i in range(self.problem.size()):
               if not i in self.assignment:
                   node = SkiSolution(self.problem, self.assignment.copy()+[i])
                   nodes.append(node)

        return nodes


    def lower_bound(self) -> float:
        bound = self.eval() # Decisions already made
        for i in range(self.size(), self.problem.size()):
            bound += abs(min(self.problem.lengths, key = lambda p: abs(p - self.problem.heights[i])) - self.problem.heights[i])
        return bound

    def estimation(self) -> float:
        return self.lower_bound()


    def greedy(self) -> search.Solution:
        assignment = self.assignment.copy()
        for i in range(self.size(), self.problem.size()):
            best = float("inf")
            best_assignment = None
            for j in range(self.problem.size()):
                if not j in assignment:
                   cost = abs(self.problem.lengths[j] - self.problem.heights[i])
                   if cost<best:
                       best = cost
                       best_assignment = j
            assignment += [best_assignment]
        return SkiSolution(self.problem, assignment)

    def upper_bound(self) -> float:
        greedy_solution = self.greedy()
        return greedy_solution.eval()

    def upper_bound_solution(self) -> (float, search.Solution):
        greedy_solution = self.greedy()
        return (greedy_solution.eval(), greedy_solution)


# Main program

if __name__=='__main__':
    problem = SkiProblem([178,168,190,170],[183,188,168,175])
    print("PROBLEM")
    print(problem)

    root = SkiSolution(problem)

    print("GREEDY")
    print(root.greedy())

    print("BRANCH & BOUND")
    (c,s) = search.branch_and_bound_min(root)
    print(f"Best cost = {c}")
    print(f"Best solution = {s}")

    print("BRANCH & BOUND (without greedy algorithm)")
    (c,s) = search.branch_and_bound_min_no_greedy(root)
    print(f"Best cost = {c}")
    print(f"Best solution = {s}")
