import unittest
import search

from ski import SkiProblem, SkiSolution

class TestSkies(unittest.TestCase):

    def setUp(self):
        self.problem = SkiProblem([178,168,190,170],[183,188,168,175])

    def test_problem(self):
        self.assertEqual(4, self.problem.size())
        self.assertEqual([178,168,190,170], self.problem.heights)
        self.assertEqual([183,188,168,175], self.problem.lengths)

    def test_root(self):
        root = SkiSolution(self.problem)
        self.assertEqual(0, root.size())
        self.assertFalse(root.solution())
        self.assertEqual(0, root.eval())
        self.assertEqual(7, root.lower_bound())
        self.assertEqual(18, root.upper_bound())

    def test_root_children(self):
        root = SkiSolution(self.problem)
        children = root.children()
        self.assertEqual(4, len(children))

        self.assertEqual(1, children[0].size())
        self.assertEqual(0, children[0].assignment[0])
        self.assertEqual(5, children[0].eval())

        self.assertEqual(1, children[1].size())
        self.assertEqual(1, children[1].assignment[0])
        self.assertEqual(10, children[1].eval())
        
        self.assertEqual(1, children[2].size())
        self.assertEqual(2, children[2].assignment[0])
        self.assertEqual(10, children[2].eval())
        
        self.assertEqual(1, children[3].size())
        self.assertEqual(3, children[3].assignment[0])
        self.assertEqual(3, children[3].eval())

    def test_partial_solution(self):
        assignment = SkiSolution(self.problem, [0])
        self.assertEqual(1, assignment.size())
        self.assertFalse(assignment.solution())
        children = assignment.children() # -> [0,1] [0,2] [0,3]
        self.assertEqual(3, len(children))
        self.assertEqual(2, children[0].size())
        self.assertEqual(0, children[0].assignment[0])
        self.assertEqual(1, children[0].assignment[1])
        self.assertEqual(2, children[1].size())
        self.assertEqual(0, children[1].assignment[0])
        self.assertEqual(2, children[1].assignment[1])
        self.assertEqual(2, children[2].size())
        self.assertEqual(0, children[2].assignment[0])
        self.assertEqual(3, children[2].assignment[1])

        assignment = SkiSolution(self.problem, [1])
        self.assertEqual(1, assignment.size())
        self.assertFalse(assignment.solution())
        children = assignment.children() # -> [1,0] [1,2] [1,3]
        self.assertEqual(3, len(children))
        self.assertEqual(2, children[0].size())
        self.assertEqual(1, children[0].assignment[0])
        self.assertEqual(0, children[0].assignment[1])
        self.assertEqual(2, children[1].size())
        self.assertEqual(1, children[1].assignment[0])
        self.assertEqual(2, children[1].assignment[1])
        self.assertEqual(2, children[2].size())
        self.assertEqual(1, children[2].assignment[0])
        self.assertEqual(3, children[2].assignment[1])


    def test_solution_suboptimal(self):
        solution = SkiSolution(self.problem, [0,1,2,3])
        self.assertEqual(4, solution.size())
        self.assertTrue(solution.solution())
        self.assertEqual(52, solution.eval())
        self.assertEqual(52, solution.lower_bound())
        self.assertEqual(52, solution.upper_bound())

    def test_solution_optimal(self):
        solution = SkiSolution(self.problem, [0,2,1,3])
        self.assertEqual(4, solution.size())
        self.assertTrue(solution.solution())
        self.assertEqual(12, solution.eval())
        self.assertEqual(12, solution.lower_bound())
        self.assertEqual(12, solution.upper_bound())

    def test_solution_greedy(self):
        root = SkiSolution(self.problem)
        greedy = root.greedy()
        self.assertEqual(4, greedy.size())
        self.assertTrue(greedy.solution())
        self.assertEqual(18, greedy.eval())
        self.assertEqual(18, greedy.lower_bound())
        self.assertEqual(18, greedy.upper_bound())
        self.assertEqual(3, greedy.assignment[0])
        self.assertEqual(2, greedy.assignment[1])
        self.assertEqual(1, greedy.assignment[2])
        self.assertEqual(0, greedy.assignment[3])

    def test_solution_branch_and_bound(self):
        root = SkiSolution(self.problem)
        (c,s) = search.branch_and_bound_min(root)
        self.assertEqual(12, c)
        self.assertEqual(4, s.size())
        self.assertTrue(s.solution())
        self.assertEqual(12, s.eval())
        self.assertEqual(12, s.lower_bound())
        self.assertEqual(12, s.upper_bound())
        self.assertEqual(0, s.assignment[0])
        self.assertEqual(2, s.assignment[1])
        self.assertEqual(1, s.assignment[2])
        self.assertEqual(3, s.assignment[3])


if __name__ == '__main__':
    unittest.main()

 